/*
    This program is for testing the CTRL-C behaviour of the 
    different console types: windows cmd shell, bash, windows 4nt shell
    
    Currently bash seems to have a problem. The CTRL-C is always handled as 
    a CTRL-BREAK there, i.e. the code is not further executed, the app 
    stops after executing the signal handler.
*/

#include <windows.h>
#include <stdio.h>
#include <signal.h>

bool loop = true;
long lClPid = 0;
HANDLE hClProcess = 0;

void setHandler();

extern "C" void signalHandler(int sig)
{
   switch( sig )
   {
      case SIGINT:  // == 2
         setHandler();
         printf("\n>>> signals: Signal SIGINT=%d received\n",sig);
         break;
      case SIGBREAK: // == 21 
         printf("\n>>> signals: Signal SIGBREAK=%d received\n",sig);
         break;
      default:
         printf("\n>>> signals: Unknown signal=%d received\n",sig);
         break;
   };
   loop=false;
   // By installing this handler the system should 
   // ignore the default action...
}

/**
 * The handlers are treated equally by the bash.
 * If more than one handler is registered then the one
 * that has been registered at last will be called at 
 * first.
 */
void setHandler()
{
   void* mode = 0;
   mode = signal( SIGINT , signalHandler);
   if (mode == SIG_ERR)
   {
      printf("\n>>> signals: Error installing signal handler for INT\n");
      return;
   }
   mode = signal( SIGBREAK, signalHandler );
   if (mode == SIG_ERR)
   {
      printf("\n>>> signals: Error installing signal handler for BREAK\n");
      return;
   }
}

int main(int argc, char* argv[])
{
   setHandler();
   
   char *pszClCommandLine = new char[512];
   memset(pszClCommandLine, '\0', 512);

   // CPP
   if(argc!=2)
   {
      printf(">>> signals: Wrong number of arguments!\n");
      printf(">>> signals: signals.exe [cpp|java]\n");
      exit(1);
   }
   if(strcmp(argv[1], "cpp")==0)
      sprintf(pszClCommandLine, "myclass.exe");
   else
   {
      // JAVA
      if(strcmp(argv[1], "java")==0)
      {
         // Avoid the waste of signal handlers by the Java VM: -Xrs
         //sprintf(pszClCommandLine, "c:\\tools32\\jdk\\1.3.1_02\\bin\\java.exe -cp . myclass");
         sprintf(pszClCommandLine, "c:\\tools32\\jdk\\1.3.1_02\\bin\\java.exe -Xrs -cp . myclass");
      }
      else
      {
         printf(">>> signals: Wrong parameter %s\n",argv[1]);
         exit(1);
      }
   }
   
   PROCESS_INFORMATION piProcInfo;
   STARTUPINFO siStartInfo;
   memset(&siStartInfo, '\0', sizeof(siStartInfo));
   memset(&piProcInfo, '\0', sizeof(piProcInfo));

   // Set up members of STARTUPINFO structure
   siStartInfo.cb = sizeof(STARTUPINFO);
   siStartInfo.lpReserved  = NULL;
   siStartInfo.lpReserved2 = NULL;
   siStartInfo.cbReserved2 = 0;
   siStartInfo.lpDesktop   = NULL;
   siStartInfo.dwFlags     = 0;

   memset(&piProcInfo, 0, sizeof(PROCESS_INFORMATION));

   // When detaching the process from the console the new process has no console
   // window and 
   //   DWORD dwCreationFlags = CREATE_NEW_PROCESS_GROUP | DETACHED_PROCESS;
   DWORD dwCreationFlags = CREATE_NEW_PROCESS_GROUP;

   if (!CreateProcess(NULL,
                      pszClCommandLine,    // call command line
                      NULL,                // process security attributes
                      NULL,                // security attributes
                      TRUE,                // handles are inherited
                      dwCreationFlags,     // creation flags
                      NULL,                // use parent's environment
                      NULL,                // use parent's current directory
                      &siStartInfo,        // STARTUPINFO pointer
                      &piProcInfo))        // receives PROCESS_INFORMATION
   {
      printf(">>> signals: Error creating process...\n");
      return -1;
   }

   lClPid     = piProcInfo.dwProcessId;
   hClProcess = piProcInfo.hProcess;

   printf("\n>>> signals: Waiting now...\n");

   while(loop)
     Sleep ((DWORD) 1000) ;

   printf("\n>>> signals: Finished waiting now...\n");

   // Windows NT/2000/XP: When a process is created with CREATE_NEW_PROCESS_GROUP specified, 
   // an implicit call to SetConsoleCtrlHandler(NULL,TRUE) is made on behalf of the new process; 
   // this means that the new process has CTRL+C disabled. This lets good shells handle CTRL+C 
   // themselves, and selectively pass that signal on to sub-processes. CTRL+BREAK is not disabled, 
   // and may be used to interrupt the process/process group.
   printf("\n>>> signals: Forwarding CTRL_BREAK_EVENT to the child process...\n");
   if(GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT,(DWORD) lClPid)==FALSE)      // process group id
   {
      printf(">>> signals: Error sending child the CTRL_BREAK_EVENT...\n");
      return -1;
   }
   
   if (NULL != piProcInfo.hThread)
      CloseHandle(piProcInfo.hThread);
   CloseHandle(hClProcess);

   return 0;
}

